#!/usr/bin/python

#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2013
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#

#import time, sys, os, sched             # sched lets you call os to schedule future commands
#
#import dobackup
#                                        #  Google or Python Coockbook page 133

#class backupinfo:
#    frequency=""
#    fullinc=""
#    bieccmap=""
#    nextid=""
#    nexttime=""
#
#def backupmanager():
#    dobackup.dobackup()                 # right now we just do one backup - not so clever at the moment





